/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.appwithinminutes.test.ui;

import java.util.Arrays;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Order;
import org.junit.jupiter.api.Test;
import org.xwiki.appwithinminutes.test.po.AppWithinMinutesHomePage;
import org.xwiki.appwithinminutes.test.po.ApplicationClassEditPage;
import org.xwiki.appwithinminutes.test.po.ApplicationCreatePage;
import org.xwiki.appwithinminutes.test.po.ApplicationHomeEditPage;
import org.xwiki.appwithinminutes.test.po.ApplicationHomePage;
import org.xwiki.appwithinminutes.test.po.ApplicationTemplateProviderEditPage;
import org.xwiki.appwithinminutes.test.po.ApplicationsLiveTableElement;
import org.xwiki.appwithinminutes.test.po.ClassFieldEditPane;
import org.xwiki.appwithinminutes.test.po.EntryEditPage;
import org.xwiki.appwithinminutes.test.po.EntryNamePane;
import org.xwiki.flamingo.skin.test.po.ChildrenPage;
import org.xwiki.index.tree.test.po.DocumentPickerModal;
import org.xwiki.livedata.test.po.TableLayoutElement;
import org.xwiki.test.docker.junit5.TestReference;
import org.xwiki.test.docker.junit5.UITest;
import org.xwiki.test.integration.junit.LogCaptureConfiguration;
import org.xwiki.test.ui.TestUtils;
import org.xwiki.test.ui.XWikiWebDriver;
import org.xwiki.test.ui.po.LiveTableElement;
import org.xwiki.test.ui.po.ViewPage;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.xwiki.appwithinminutes.test.po.ApplicationClassEditPage.EMPTY_CANVAS_HINT;
import static org.xwiki.appwithinminutes.test.po.ApplicationCreatePage.APP_NAME_USED_WARNING_MESSAGE;
import static org.xwiki.flamingo.skin.test.po.ChildrenPage.LIVE_DATA_TITLE;

/**
 * Tests the application wizard.
 *
 * @version $Id: 36c185e2f1a0fa3754ffecccd4b0afe9002bcb81 $
 * @since 12.5
 */
@UITest
class WizardIT
{
    private static final String USER_NAME = "Wizard";

    private static final String PASSWORD = "password";

    @BeforeAll
    static void beforeAll(TestUtils testUtils)
    {
        testUtils.loginAsSuperAdmin();
        // The application creator needs script rights in order to execute the scripts generated by the wizard.
        testUtils.setGlobalRights("", "XWiki." + USER_NAME, "script", true);
        testUtils.createUser(USER_NAME, PASSWORD, "");
    }

    @Test
    @Order(1)
    void createApplicationWhenDefaultWikiSyntaxIsPlain(TestReference testReference, TestUtils testUtils,
        XWikiWebDriver driver) throws Exception
    {
        testUtils.loginAsSuperAdmin();
        String defaultDocSyntax = testUtils.setWikiPreference("core.defaultDocumentSyntax", "plain/1.0");

        String appName = testReference.getLastSpaceReference().getName();
        AppWithinMinutesHomePage.gotoPage().deleteApplication(appName);

        testUtils.login(USER_NAME, PASSWORD);

        try {
            ApplicationCreatePage appCreatePage = AppWithinMinutesHomePage.gotoPage().clickCreateApplication();
            appCreatePage.setApplicationName(appName);
            ApplicationClassEditPage classEditPage = appCreatePage.clickNextStep();
            classEditPage.addField("Short Text");

            ApplicationHomeEditPage appHomeEditPage = classEditPage.clickNextStep().clickNextStep();
            appHomeEditPage.setDescription("one **two** three");

            ApplicationHomePage appHomePage = appHomeEditPage.clickFinish();
            // Verify that the application home page was created with the right syntax.
            assertTrue(driver.getPageSource().contains("one <strong>two</strong> three"),
                "The application description is not rendered properly on the application home page.");

            EntryNamePane entryNamePage = appHomePage.clickAddNewEntry();
            entryNamePage.setName("Test");

            EntryEditPage entryEditPage = entryNamePage.clickAdd();
            String value = "0123456789";
            entryEditPage.setValue("shortText1", value);

            ViewPage entryViewPage = entryEditPage.clickSaveAndView();
            // Verify that the application sheet was created with the right syntax.
            assertTrue(entryViewPage.getContent().contains(value),
                "The entry view page doesn't display the property value.");
        } finally {
            // Restore the default document syntax.
            testUtils.loginAsSuperAdmin();
            testUtils.setWikiPreference("core.defaultDocumentSyntax", defaultDocSyntax);
        }
    }

    /**
     * Tests the application creation process from start to end.
     * @since 13.3RC1
     * @since 12.10.6
     */
    @Test
    @Order(2)
    void createApplication(TestUtils testUtils, TestReference testReference,
        LogCaptureConfiguration logCaptureConfiguration)
    {
        ApplicationCreatePage appCreatePage = goToAppCreatePage(testUtils, testReference);

        // Step 1
        // Set the application location.
        appCreatePage.getLocationPicker().browseDocuments();
        new DocumentPickerModal().selectDocument(getClass().getSimpleName(),
            testReference.getLastSpaceReference().getName(), "WebHome");
        appCreatePage.getLocationPicker().waitForLocation(
            Arrays.asList("", getClass().getSimpleName(), testReference.getLastSpaceReference().getName(), ""));

        // Enter the application name, making sure we also use some special chars.
        // See XWIKI-11747: Impossible to create new entry with an application having UTF8 chars in its name
        String appName = "Cities âé";
        String[] appPath = new String[] { getClass().getSimpleName(), testReference.getLastSpaceReference().getName(),
            appName };
        appCreatePage.setApplicationName(appName);

        // Move to the next step.
        ApplicationClassEditPage classEditPage = appCreatePage.clickNextStep();

        // Step 2
        // Add a 'Short Text' field.
        ClassFieldEditPane fieldEditPane = classEditPage.addField("Short Text");

        // Set the field pretty name and default value
        fieldEditPane.setPrettyName("City Name");
        fieldEditPane.setDefaultValue("Paris");

        // Move to the next step.
        ApplicationTemplateProviderEditPage templateProviderEditPage = classEditPage.clickNextStep();

        // Move back to the second step.
        classEditPage = templateProviderEditPage.clickPreviousStep();

        // Open the configuration panel and set the field name
        fieldEditPane = new ClassFieldEditPane("shortText1");
        fieldEditPane.openConfigPanel();
        fieldEditPane.setName("cityName");

        // Move to the next step.
        templateProviderEditPage = classEditPage.clickNextStep();

        // Step 3
        templateProviderEditPage.setIcon("worl");
        templateProviderEditPage.setDescription("A city page");

        // Move to the next step.
        ApplicationHomeEditPage homeEditPage = templateProviderEditPage.clickNextStep();

        // Step 4
        // Enter the application description.
        String appDescription = "Simple application to manage data about various cities";
        homeEditPage.setDescription(appDescription);

        // Add the Short Text field from the previous step to the list of columns.
        homeEditPage.addLiveTableColumn("City Name");

        // Click the finish button which should lead us to the application home page.
        ApplicationHomePage homePage = homeEditPage.clickFinish();

        // Assert the application description is present.
        assertTrue(homePage.getContent().contains(appDescription));

        // Add a new entry.
        String firstEntryName = "City 1 âé";
        EntryNamePane entryNamePane = homePage.clickAddNewEntry();
        entryNamePane.setName(firstEntryName);
        EntryEditPage entryEditPage = entryNamePane.clickAdd();

        // Assert the pretty name and the default value of the Short Text field.
        // Apparently WebElement#getText() takes into account the text-transform CSS property.
        assertEquals("CITY NAME", entryEditPage.getLabel("cityName"));
        assertEquals("Paris", entryEditPage.getValue("cityName"));

        // Change the field value.
        entryEditPage.setValue("cityName", "London");

        // Save and go back to the application home page.
        entryEditPage.clickSaveAndView().clickBreadcrumbLink(appName);
        homePage = new ApplicationHomePage();

        // Assert the entry we have just created is listed in the live table.
        LiveTableElement entriesLiveTable = homePage.getEntriesLiveTable();
        entriesLiveTable.waitUntilReady();
        assertTrue(entriesLiveTable.hasRow("City Name", "London"));

        // Assert that only the entry we have just created is listed as child of the application home page. The rest of
        // the documents (class, template, sheet, preferences) should be marked as hidden.
        TableLayoutElement childrenTableLayout = ChildrenPage.clickOnChildrenMenu().getLiveData().getTableLayout();
        assertEquals(1, childrenTableLayout.countRows());
        childrenTableLayout.assertRow(LIVE_DATA_TITLE, firstEntryName);

        // Go back to the application home edit page.
        testUtils.gotoPage(
            Arrays.asList(getClass().getSimpleName(), testReference.getLastSpaceReference().getName(), appName),
            "WebHome", "edit", "");
        homeEditPage = new ApplicationHomeEditPage();

        // Change the application description.
        appDescription = "The best app!";
        homeEditPage.setDescription(appDescription);

        // Remove one of the live table columns.
        homeEditPage.removeLiveTableColumn("Actions");

        // Save
        homePage = homeEditPage.clickSaveAndView();

        // Assert that the application description has changed and that the column has been removed.
        assertTrue(homePage.getContent().contains(appDescription));
        entriesLiveTable = homePage.getEntriesLiveTable();
        entriesLiveTable.waitUntilReady();
        assertFalse(entriesLiveTable.hasColumn("Actions"));

        // Click the link to edit the application.
        classEditPage = homePage.clickEditApplication();

        // Drag a Number field.
        fieldEditPane = classEditPage.addField("Number");

        // Set the field pretty name.
        fieldEditPane.setPrettyName("Population Size");

        // Fast forward.
        homeEditPage = classEditPage.clickNextStep().clickNextStep();
        // Just wait for the WYSIWYG editor (which is used for setting the application description) to load so that the
        // page layout is stable before we click on the Finish button.
        homeEditPage.setDescription(appDescription);
        homePage = homeEditPage.clickFinish();

        // Add a new entry.
        String secondEntryName = "City 2 âé";
        entryNamePane = homePage.clickAddNewEntry();
        entryNamePane.setName(secondEntryName);
        entryEditPage = entryNamePane.clickAdd();

        // Assert the new field is displayed in the edit sheet (field name was auto-generated).
        // Apparently WebElement#getText() takes into account the text-transform CSS property.
        assertEquals("POPULATION SIZE", entryEditPage.getLabel("number1"));

        // Save and go back to the application home page.
        entryEditPage.clickSaveAndView().clickBreadcrumbLink(appName);
        homePage = new ApplicationHomePage();

        // Assert both entries are displayed in the live table.
        entriesLiveTable = homePage.getEntriesLiveTable();
        entriesLiveTable.waitUntilReady();
        assertTrue(entriesLiveTable.hasRow("Page Title", firstEntryName));
        assertTrue(entriesLiveTable.hasRow("Page Title", secondEntryName));

        // Go to the App Within Minutes home page.
        AppWithinMinutesHomePage appWithinMinutesHomePage = AppWithinMinutesHomePage.gotoPage();

        // Assert that the created application is listed in the live table.
        ApplicationsLiveTableElement appsLiveTable = appWithinMinutesHomePage.getAppsLiveTable();
        assertTrue(appsLiveTable.isApplicationListed(appName));

        // Delete the application entries.
        homePage = appsLiveTable.viewApplication(appName);
        assertEquals('/' + StringUtils.join(appPath, '/'), homePage.getBreadcrumbContent());
        homePage.clickDeleteAllEntries().clickYes();
        // Verify that the entries live table is empty.
        entriesLiveTable = homePage.getEntriesLiveTable();
        entriesLiveTable.waitUntilReady();
        assertEquals(0, entriesLiveTable.getRowCount());

        // Delete the application.
        homePage.clickDeleteApplication().clickYes();
        // Verify that the application is not listed anymore.
        appsLiveTable = AppWithinMinutesHomePage.gotoPage().getAppsLiveTable();
        assertFalse(appsLiveTable.isApplicationListed(appName));

        logCaptureConfiguration.registerExcludes(
            "WikiComponentException: Registering UI extensions at wiki level requires wiki administration rights");
    }

    /**
     * @see "XWIKI-7380: Cannot go back from step 2 to step 1"
     * @since 13.3RC1
     * @since 12.10.6
     */
    @Test
    @Order(3)
    void goBackToFirstStep(TestUtils testUtils, TestReference testReference)
    {
        ApplicationCreatePage appCreatePage = goToAppCreatePage(testUtils, testReference);

        // Step 1
        String appName = "Empty App âé";
        appCreatePage.setApplicationName(appName);

        // Step 2
        ApplicationClassEditPage classEditPage = appCreatePage.clickNextStep();
        classEditPage.addField("Short Text");

        // Back to Step 1
        appCreatePage = classEditPage.clickPreviousStep();
        appCreatePage.setApplicationName(appName);
        // Test that the application wasn't created.
        assertFalse(appCreatePage.getContent().contains(APP_NAME_USED_WARNING_MESSAGE));

        // Step 2 again
        classEditPage = appCreatePage.clickNextStep();
        assertTrue(classEditPage.getContent().contains(EMPTY_CANVAS_HINT));
        classEditPage.addField("Number");

        // Step 4 and back to Step 2
        classEditPage = classEditPage.clickNextStep().clickNextStep().clickPreviousStep().clickPreviousStep();
        assertFalse(classEditPage.getContent().contains(EMPTY_CANVAS_HINT));
        assertFalse(classEditPage.hasPreviousStep());
    }

    private ApplicationCreatePage goToAppCreatePage(TestUtils testUtils, TestReference testReference)
    {
        // Register a simple user, login and go to the App Within Minutes home page.
        String userName = "SimpleUser";
        String password = "SimplePassword";
        testUtils.createUserAndLogin(userName, password);
        // Make sure the application location exists so that we can select it with the location picker.
        testUtils.createPage(Arrays.asList(getClass().getSimpleName(), testReference.getLastSpaceReference().getName()),
            "WebHome", null, null);
        AppWithinMinutesHomePage appWithinMinutesHomePage = AppWithinMinutesHomePage.gotoPage();

        // Click the Create Application button.
        return appWithinMinutesHomePage.clickCreateApplication();
    }
}
